/***************************************************************/
/*  Function fft2(complex *X, int M)                           */
/*                                                             */
/*   This is an elementary, complex, radix 2, decimation in    */
/*  time FFT. The computations are performed "in place" and    */
/*  the output overwrites the input array.  This FFT is faster */
/*  than fft1() because W**k is read from a stored sine table  */
/*  rather than computed recursively. On the other hand, it    */
/*  requires more memory to store the sine table.              */
/*                                                             */
/*  The sine table should be a global array called sinetable[].*/
/***************************************************************/

#include "complex.h"/* Definition of complex variable structure */
#include <math.h>   /* Definitions for math library             */

extern float sinetable[];

void fft(struct cmpx *X, int M)
         /* X is an array of N = 2**M complex points. */
  {
  struct cmpx temp1;          /* temporary storage complex variable */
  struct cmpx WK;             /* Twiddle factor W**k                */
  int i,j,k;              /* loop indexes                       */
  int id;                 /* Index of lower point in butterfly  */
  int N = 1 << M;         /* Number of points for FFT           */
  int N2 = N/2;
  int L;                  /* FFT stage                          */
  int LE;           /* Number of points in sub DFT at stage L,  */
                    /* and offset to next DFT in stage          */
  int LE1;                /* Number of butterflys in one DFT at */
                          /*  stage L. Also is offset to lower  */
                          /*  point in butterfly at stage L     */
  int w_index;            /*  sinetable index for W**k          */
  int w_inc;              /*  w_inc = N/LE = angle inc for W**k */
  int cos_offset = N/4;   /*  offset to cosine in sinetable[]   */

/*==============================================================*/
/*  Rearrange input array in bit-reversed order                 */
/*                                                              */
/*    The index j is the bit reversed value of i.  Since 0 -> 0 */
/*  and N-1 -> N-1 under bit-reversal, these two reversals are  */
/*  skipped.                                                    */

  j = 0;
  for(i=1; i<(N-1); i++)
    {

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*  Increment bit-reversed counter for j by adding 1 to msb and */
/*   propagating carries from left to right.                    */
                                            
    k = N2;               /* k is 1 in msb, 0 elsewhere         */

/*--------------------------------------------------------------*/
/*  Propagate carry from left to right                          */

    while(k<=j)           /* Propagate carry if bit is 1        */
      {
      j = j - k;          /* Bit tested is 1, so clear it.      */
      k = k/2;            /* Set up 1 for next bit to right.    */
      }
    j = j+k;              /* Change 1st 0 from left to 1        */
/*--------------------------------------------------------------*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/*  Swap samples at locations i and j if not previously swapped.*/

    if(i<j)               /* Test if previously swapped.        */
      {
      temp1.real = X[j].real;
      temp1.imag = X[j].imag;
      X[j].real = X[i].real;
      X[j].imag = X[i].imag;
      X[i].real = temp1.real;
      X[i].imag = temp1.imag;
      }
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

    }

/*==============================================================*/
/*  Do M stages of butterflys                                   */

  for(L=1; L<= M; L++)
    {
    LE = 1 << L;            /*  LE = 2**L = points in sub DFT   */
    LE1 = LE/2;             /* Number of butterflys in sub-DFT  */
    w_inc = N/LE;  /* sinetable[] increment for W**k at stage L */
/*--------------------------------------------------------------*/
/*   Do butterflys for L-th stage                               */
    	
    	/* Initialize W**k as W**0 = 1 + j0                     */
       	w_index = 0;        /* k*(N/LE) for k = 0               */
        WK.real = 1.0;      /* cos(0) = sinetable[cos_offset]   */
        WK.imag = 0.0;      /* -sin(0) = sinetable[0]           */                             

    for(j=0; j<LE1; j++)    /* Do the LE1 butterflys per sub DFT*/
      {
/*..............................................................*/
/*    Compute butterflys that use same W**k                     */
 
      for(i=j; i<N; i += LE)
        {
        id = i + LE1;      /* Index of lower point in butterfly */
        temp1.real = X[id].real*WK.real - X[id].imag*WK.imag;
        temp1.imag = X[id].imag*WK.real + X[id].real*WK.imag;

        X[id].real = X[i].real - temp1.real;
        X[id].imag = X[i].imag - temp1.imag;

        X[i].real  = X[i].real + temp1.real;
        X[i].imag  = X[i].imag + temp1.imag;
        }
/*..............................................................*/
/* W**(k+1)=sinetable[(k+1)*N/LE + N/4] - j sinetable[(k+1)*N/LE] */

      w_index += w_inc;
      WK.real = sinetable[w_index + cos_offset];
      WK.imag = -sinetable[w_index];
/*..............................................................*/
      }
/*--------------------------------------------------------------*/
    }
  return;
  }
